from __future__ import annotations

import logging
from typing import Any, Callable, Dict, List, Optional

from openai import AzureOpenAI, OpenAI

from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE


class ActionException(Exception):
    pass


class Action:
    def __init__(
        self,
        name,
        function,
        pydantic_model,
        stop=False,  # TODO: move all `stop` argument, self.orch in chat.completions.create
        decorators: List[Callable[..., None]] = [],
        description=None,
        logger=None,
        logging_metadata: Optional[dict] = None,
        logging_level=logging.INFO,
    ):
        self.name = name
        self.logger = logger
        self.stop = stop
        self.decorators = decorators

        if function.__doc__ is None and description is None:
            raise ActionException(
                f"Decorated method under action {name} must have a docstring for description."
            )
        self.description = description or function.__doc__

        self.pydantic_model = pydantic_model

        self.undecorated_function = function
        for decorator in self.decorators:
            function = decorator(function)
        self.function = function

        if self.logger:
            self.function = traceable(
                self.name,
                self.logger,
                metadata=logging_metadata,
                level=logging_level,
            )(self.function)

        self.__module__ = self.function.__module__
        self.__name__ = self.function.__name__
        self.__qualname__ = self.function.__qualname__
        self.__annotations__ = self.function.__annotations__
        self.__doc__ = self.function.__doc__

    def json_schema(self):
        return self.pydantic_model.model_json_schema()

    def invoke(
        self,
        client=None,
        force=True,
        logger=None,
        token_usage_tracker=None,
        *args,
        **kwargs,
    ):
        from actionweaver.llms.wrapper import ActionWeaverLLMClientWrapper

        if type(client) in (OpenAI, AzureOpenAI):
            return client.chat.completions.create(
                actions=[self],
                orch={DEFAULT_ACTION_SCOPE: self, self.name: None} if force else None,
                logger=logger,
                token_usage_tracker=token_usage_tracker,
                *args,
                **kwargs,
            )
        elif type(client) == ActionWeaverLLMClientWrapper:
            return client.create(
                actions=[self],
                orch={DEFAULT_ACTION_SCOPE: self, self.name: None} if force else None,
                logger=logger,
                token_usage_tracker=token_usage_tracker,
                *args,
                **kwargs,
            )
        else:
            raise ActionException(
                f"Client type {type(client)} not supported in invoke method. Please use OpenAI or AzureOpenAI client."
            )

    def get_function_details(self):
        return {
            "name": self.name,
            "description": self.description,
            "parameters": self.json_schema(),
        }

    def bind(self, instance) -> InstanceAction:
        return InstanceAction(
            self.name,
            self.function,
            self.pydantic_model,
            self.logger,
            self.stop,
            instance=instance,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response = self.function(*args, **kwargs)

        return response

    def __get__(self, instance, owner) -> InstanceAction:
        """

        Note:
            The `__get__` method is a descriptor method that is called when the action is accessed from an instance.
            It returns an instance-specific action method that is bound to the given instance.
        """
        return InstanceAction(
            self.name,
            self.function,
            self.pydantic_model,
            self.logger,
            self.stop,
            instance=instance,
        )

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __eq__(self, other):
        return isinstance(other, Action) and self.name == other.name

    def __str__(self):
        return self.name


class InstanceAction(Action):
    def __init__(
        self,
        name,
        function,
        pydantic_model,
        logger=None,
        stop=False,
        instance=None,
    ):
        super().__init__(name, function, pydantic_model, stop=stop, logger=logger)
        self.instance = instance
        self.pydantic_model = pydantic_model

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response = self.function(self.instance, *args, **kwargs)

        return response


class ActionHandlers:
    def __init__(self, *args, **kwargs):
        self.name_to_action: Dict[str, Action] = {}

    def contains(self, name) -> bool:
        return name in self.name_to_action

    @classmethod
    def from_actions(cls, actions):
        ret = cls()
        ret.name_to_action = {action.name: action for action in actions}
        return ret

    def __len__(self) -> int:
        return len(self.name_to_action)

    def check_orchestration_expr_validity(self, expr):
        if expr is None:
            return

        if isinstance(expr, str):
            if expr not in self.name_to_action:
                raise ActionException(f"Action {expr} not found.")
            return

        for element in expr:
            self.check_orchestration_expr_validity(element)

    @classmethod
    def merge(cls, *handlers) -> ActionHandlers:
        merged = cls()
        for handler in handlers:
            merged.name_to_action.update(handler.name_to_action)
        return merged

    def __getitem__(self, key) -> Action:
        return self.name_to_action[key]



from local_llm_function_calling import Generator

# understand how constraint works
# use new model

#

functions = [
    {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                    "maxLength": 20,
                },
                "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
            },
            "required": ["location"],
        },
    }
]

# Initialize the generator with the Hugging Face model and our functions
generator = Generator.hf(functions, "gpt2")

# Generate text using a prompt
function_call = generator.generate("What is the weather like today in Brooklyn?")
print(function_call)



import logging
from typing import Any, Callable, Dict, List, Optional

from actionweaver.actions.action import Action, ActionException
from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.pydantic_utils import create_pydantic_model_from_func


def create_pydantic_model_from_function(
    function,
    override_params=None,  # override_params: Optional dictionary of parameters to override kwarg and non-kwarg of decorated method.
):

    return create_pydantic_model_from_func(
        function.__name__.title(),
        function,
        override_params=override_params,
    )


def action(
    name,
    pydantic_model=None,
    logger=None,
    stop=False,
    description=None,
    decorators: List[Callable[..., None]] = [],
    logging_metadata: Optional[dict] = None,
    logging_level=logging.INFO,
):

    _logger = logger

    def create_action(function):
        return Action(
            name=name,
            function=function,
            pydantic_model=(
                pydantic_model
                if pydantic_model
                else create_pydantic_model_from_func(
                    function.__name__.title(),
                    function,
                )
            ),
            stop=stop,
            decorators=decorators,
            description=description,
            logger=_logger,
            logging_metadata=logging_metadata,
            logging_level=logging_level,
        )

    return create_action



from __future__ import annotations

import itertools
import json
import logging
import time
import uuid
from itertools import chain
from typing import List, Optional

from openai import Stream
from openai.types.chat.chat_completion_message import FunctionCall

import actionweaver.llms.loop_action as la
from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.azure.functions import Functions
from actionweaver.llms.exception_handler import ChatLoopInfo, ExceptionHandler
from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator
from actionweaver.utils.tokens import TokenUsageTracker

# TODO: support AsyncAzureOpenAI


class FunctionCallingLoopException(Exception):
    def __init__(self, message="", extra_info=None):
        super().__init__(message)
        self.extra_info = extra_info or {}

    def __str__(self):
        # Customize the string representation to include extra_info
        extra_info_str = ", ".join(
            f"{key}: {value}" for key, value in self.extra_info.items()
        )
        return f"{super().__str__()} | Additional Info: [{extra_info_str}]"


def handle_stream_response(api_response):
    _, iterator = get_first_element_and_iterator(api_response)
    stream_function_calls = None
    for chunk in iterator:
        if chunk.choices:
            delta = chunk.choices[0].delta

            if delta.function_call:
                if stream_function_calls is None:
                    # if function call detected, we merge all deltas and treat it as the non-stream response
                    stream_function_calls = FunctionCall(name="", arguments="")

                stream_function_calls.name += (
                    delta.function_call.name if delta.function_call.name else ""
                )

                stream_function_calls.arguments += (
                    delta.function_call.arguments
                    if delta.function_call.arguments
                    else ""
                )

            elif delta.content:
                # if it has content return as generator right away
                return chain([chunk], iterator)
            else:
                raise FunctionCallingLoopException(
                    f"Unsupported streaming response",
                    extra_info={
                        "message": "Unsupported streaming response",
                        "timestamp": time.time(),
                    },
                )
    return stream_function_calls


def build_orch(actions: List[Action] = None, orch=None):
    action_handler = ActionHandlers()

    if orch is None:
        orch = {}
    if DEFAULT_ACTION_SCOPE not in orch:
        orch[DEFAULT_ACTION_SCOPE] = actions

    buf = actions + list(orch.values())
    for element in buf:
        if isinstance(element, list):
            for e in element:
                action_handler.name_to_action[e.name] = e
        elif isinstance(element, Action):
            action_handler.name_to_action[element.name] = element
    # default action scope if not following actions not specified
    for _, action in action_handler.name_to_action.items():
        if action.name not in orch:
            orch[action.name] = DEFAULT_ACTION_SCOPE

    return action_handler, orch


def invoke_function(
    messages,
    model,
    function_call,
    functions,
    orch,
    action_handler,
):
    """Invoke the function, update the messages, returns functions argument for the next OpenAI API call or halt the function loop and return the response."""

    if isinstance(function_call, FunctionCall):
        function_call = function_call.model_dump()

    messages += [
        {
            "role": "assistant",
            "content": None,
            "function_call": function_call,
        }
    ]

    name = function_call["name"]
    arguments = function_call["arguments"]

    if action_handler.contains(name):
        try:
            arguments = json.loads(function_call["arguments"])
        except json.decoder.JSONDecodeError as e:
            raise FunctionCallingLoopException(
                "Parsing function call arguments from OpenAI response failed",
                extra_info={
                    "arguments": function_call["arguments"],
                    "timestamp": time.time(),
                    "model": model,
                },
            ) from e

        # Invoke action
        function_response = action_handler[name](**arguments)
        stop = action_handler[name].stop
        messages += [
            {
                "role": "function",
                "name": name,
                "content": str(function_response),
            }
        ]

        # use tools in orch[DEFAULT_ACTION_SCOPE] if expr is DEFAULT_ACTION_SCOPE
        expr = (
            orch[name]
            if orch[name] != DEFAULT_ACTION_SCOPE
            else orch[DEFAULT_ACTION_SCOPE]
        )
        return (
            Functions.from_expr(
                expr,
            ),
            (stop, function_response),
        )
    else:
        raise FunctionCallingLoopException(
            f"{name} is not a valid function name",
            extra_info={
                "timestamp": time.time(),
                "model": model,
            },
        )


def validate_orch(orch):
    if orch is not None:
        for key in orch.keys():
            if not isinstance(key, str):
                raise FunctionCallingLoopException(
                    f"Orch keys must be action name (str), found {type(key)}"
                )


def create_chat_loop(original_create_method):
    def wrapper_for_logging(
        *args,
        logger: Optional[logging.Logger] = None,
        logging_name: Optional[str] = None,
        logging_metadata: Optional[dict] = None,
        logging_level=logging.INFO,
        exception_handler: ExceptionHandler = None,
        **kwargs,
    ):
        DEFAULT_LOGGING_NAME = "actionweaver_initial_chat_completion"

        def new_create(
            actions: List[Action] = [],
            orch=None,
            token_usage_tracker=None,
            *args,
            **kwargs,
        ):
            validate_orch(orch)

            chat_completion_create_method = original_create_method
            if logger:
                chat_completion_create_method = traceable(
                    name=(logging_name or DEFAULT_LOGGING_NAME)
                    + ".chat.completions.create",
                    logger=logger,
                    metadata=logging_metadata,
                    level=logging_level,
                )(original_create_method)

            if token_usage_tracker is None:
                token_usage_tracker = TokenUsageTracker()

            argument_check(*args, **kwargs)

            messages = kwargs.get("messages")
            model = kwargs.get("model")

            action_handler, orch = build_orch(actions, orch)

            functions = Functions.from_expr(orch[DEFAULT_ACTION_SCOPE])

            while True:

                function_argument = functions.to_arguments()
                chat_loop_action = la.Unknown

                try:
                    if functions:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                            **function_argument,
                        )
                    else:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                        )

                    chat_loop_action = handle_response(
                        api_response,
                        token_usage_tracker,
                        messages,
                        model,
                        functions,
                        orch,
                        action_handler,
                        logger,
                    )
                except Exception as e:

                    if exception_handler:
                        chat_loop_action = exception_handler.handle_exception(
                            e,
                            ChatLoopInfo(
                                context={
                                    "response": api_response,
                                    "messages": messages,
                                    "functions": functions,
                                    "model": model,
                                    "orch": orch,
                                }
                            ),
                        )
                    else:
                        raise e

                if isinstance(chat_loop_action, la.ReturnRightAway):
                    return chat_loop_action.content
                elif isinstance(chat_loop_action, la.Continue):
                    functions = chat_loop_action.functions
                else:
                    raise FunctionCallingLoopException(
                        f"Unsupported chat loop action: {chat_loop_action}"
                    )

        if logger:
            return traceable(
                name=logging_name or DEFAULT_LOGGING_NAME,
                logger=logger,
                metadata=logging_metadata,
                level=logging_level,
            )(new_create)(*args, **kwargs)
        else:
            return new_create(*args, **kwargs)

    return wrapper_for_logging


def argument_check(
    *args,
    **kwargs,
):
    if "messages" not in kwargs:
        raise FunctionCallingLoopException(
            "messages keyword argument is required for chat completion"
        )
    if "model" not in kwargs:
        raise FunctionCallingLoopException(
            "model keyword argument is required for chat completion"
        )

    if "functions" in kwargs:
        raise FunctionCallingLoopException(
            "functions keyword argument is not allowed for this method, use actions instead"
        )

    if "function_call" in kwargs:
        raise FunctionCallingLoopException(
            "function_call keyword argument is not allowed for this method, use actions instead"
        )


def handle_response(
    api_response,
    token_usage_tracker,
    messages,
    model,
    functions,
    orch,
    action_handler,
    logger=None,
) -> la.LoopAction:
    # logic to handle streaming API response
    processed_stream_response = None
    if isinstance(api_response, Stream):
        processed_stream_response = handle_stream_response(api_response, logger)

        if type(processed_stream_response) == itertools.chain:
            # if it's a tee object, return the message right away
            return processed_stream_response
    else:
        token_usage_tracker.track_usage(api_response.usage)

    if processed_stream_response is not None:
        functions, (
            stop,
            resp,
        ) = invoke_function(
            messages,
            model,
            processed_stream_response,
            functions,
            orch,
            action_handler,
        )
        if stop:
            return la.ReturnRightAway(content=resp)
        else:
            return la.Continue(functions=functions)
    else:
        choice = api_response.choices[0]
        message = choice.message

        if message.function_call:
            functions, (
                stop,
                resp,
            ) = invoke_function(
                messages,
                model,
                message.function_call,
                functions,
                orch,
                action_handler,
            )
            if stop:
                return la.ReturnRightAway(content=resp)
            else:
                return la.Continue(functions=functions)
        elif message.content is not None:
            # ignore last message in the function loop
            # messages += [{"role": "assistant", "content": message["content"]}]
            if choice.finish_reason == "stop":
                """
                Stop Reasons:

                - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                """
                return la.ReturnRightAway(content=api_response)
        else:
            raise FunctionCallingLoopException(
                f"Unsupported response from OpenAI api: {api_response}"
            )



from actionweaver.actions import Action


class FunctionException(Exception):
    pass


class Functions:
    def __init__(self, function_call=None, functions=None) -> None:
        self.functions = functions
        self.function_call = function_call

    @classmethod
    def from_expr(cls, expr):
        if expr is None:
            return cls()
        elif isinstance(expr, Action):
            return cls(
                function_call={"name": expr.name},
                functions=[
                    {
                        "name": expr.name,
                        "description": expr.description,
                        "parameters": expr.json_schema(),
                    }
                ],
            )
        elif isinstance(expr, list):
            return cls(
                functions=[
                    {
                        "name": action.name,
                        "description": action.description,
                        "parameters": action.json_schema(),
                    }
                    for action in expr
                ],
                function_call="auto",
            )
        else:
            raise FunctionException(f"Invalid orchestration expression: {expr}")

    def to_arguments(self):
        return {
            "functions": self.functions,
            "function_call": self.function_call,
        }

    def __bool__(self):
        return bool(self.functions)



import functools
import inspect


def preserve_original_signature(decorator):
    def decorated_decorator(func):
        decorated = decorator(func)
        decorated.__signature__ = inspect.signature(
            func
        )  # Preserve the original function's signature

        return decorated

    return decorated_decorator


def cache(func):
    wrapper = functools.cache(func)
    wrapper.__signature__ = inspect.signature(
        func
    )  # Preserve the original function's signature
    return wrapper


def lru_cache(maxsize=128, typed=False):
    def decorator(func):
        wrapper = functools.lru_cache(maxsize, typed)(func)
        wrapper.__signature__ = inspect.signature(func)
        return wrapper

    return decorator



import itertools


def process_and_display_output(output, messages):
    processed_output = ""
    if type(output) == itertools._tee:
        for chunk in output:
            content = chunk["choices"][0].get("delta", {}).get("content")
            if content is not None:
                processed_output += content
                print(content, end="")
    else:
        processed_output = output
        print(processed_output)

    messages += [{"role": "assistant", "content": processed_output}]



from __future__ import annotations

import itertools
import json
import logging
import time
import uuid
from itertools import chain
from typing import List, Optional, Union

from openai import AsyncAzureOpenAI, AzureOpenAI, Stream
from openai.types.chat.chat_completion_message import (
    ChatCompletionMessage,
    FunctionCall,
)

from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.azure.functions import Functions
from actionweaver.telemetry import traceable
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import get_first_element_and_iterator, merge_dicts
from actionweaver.utils.tokens import TokenUsageTracker

# TODO: support AsyncAzureOpenAI


class ChatCompletionException(Exception):
    def __init__(self, message="", extra_info=None):
        super().__init__(message)
        self.extra_info = extra_info or {}

    def __str__(self):
        # Customize the string representation to include extra_info
        extra_info_str = ", ".join(
            f"{key}: {value}" for key, value in self.extra_info.items()
        )
        return f"{super().__str__()} | Additional Info: [{extra_info_str}]"


class ChatCompletion:
    def __init__(
        self,
        model,
        azure_endpoint,
        api_key,
        api_version,
        token_usage_tracker=None,
        logger=None,
        azure_deployment="",
    ):
        self.model = model
        self.action_handlers = ActionHandlers()
        self.logger = logger or logging.getLogger(__name__)
        if azure_deployment == "":
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint, api_key=api_key, api_version=api_version
            )
        else:
            self.client = AzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                api_version=api_version,
                azure_deployment=azure_deployment,
            )
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker()

    @staticmethod
    def _handle_stream_response(api_response):
        _, iterator = get_first_element_and_iterator(api_response)
        stream_function_calls = None
        for chunk in iterator:
            if chunk.choices:
                delta = chunk.choices[0].delta

                if delta.function_call:
                    if stream_function_calls is None:
                        # if function call detected, we merge all deltas and treat it as the non-stream response
                        stream_function_calls = FunctionCall(name="", arguments="")

                    stream_function_calls.name += (
                        delta.function_call.name if delta.function_call.name else ""
                    )

                    stream_function_calls.arguments += (
                        delta.function_call.arguments
                        if delta.function_call.arguments
                        else ""
                    )

                elif delta.content:
                    # if it has content return as generator right away
                    return chain([chunk], iterator)
                else:
                    raise ChatCompletionException(
                        f"Unsupported streaming response",
                        extra_info={
                            "message": "Unsupported streaming response",
                            "timestamp": time.time(),
                        },
                    )
        return stream_function_calls

    @staticmethod
    def build_orch(actions: List[Action] = None, orch=None):
        action_handler = ActionHandlers()

        if orch is None:
            orch = {}
        if DEFAULT_ACTION_SCOPE not in orch:
            orch[DEFAULT_ACTION_SCOPE] = actions

        buf = actions + list(orch.values())
        for element in buf:
            if isinstance(element, list):
                for e in element:
                    action_handler.name_to_action[e.name] = e
            elif isinstance(element, Action):
                action_handler.name_to_action[element.name] = element
        # default action scope if not following actions not specified
        for _, action in action_handler.name_to_action.items():
            if action.name not in orch:
                orch[action.name] = DEFAULT_ACTION_SCOPE

        return action_handler, orch

    @staticmethod
    def _invoke_function(
        messages,
        model,
        function_call,
        functions,
        orch,
        action_handler,
    ):
        """Invoke the function, update the messages, returns functions argument for the next OpenAI API call or halt the function loop and return the response."""

        if isinstance(function_call, FunctionCall):
            function_call = function_call.model_dump()

        messages += [
            {
                "role": "assistant",
                "content": None,
                "function_call": function_call,
            }
        ]

        name = function_call["name"]
        arguments = function_call["arguments"]

        if action_handler.contains(name):
            try:
                arguments = json.loads(function_call["arguments"])
            except json.decoder.JSONDecodeError as e:
                raise ChatCompletionException(
                    e,
                    extra_info={
                        "message": "Parsing function call arguments from OpenAI response ",
                        "arguments": function_call["arguments"],
                        "timestamp": time.time(),
                        "model": model,
                    },
                ) from e

            # Invoke action
            function_response = action_handler[name](**arguments)
            stop = action_handler[name].stop
            messages += [
                {
                    "role": "function",
                    "name": name,
                    "content": str(function_response),
                }
            ]

            # use tools in orch[DEFAULT_ACTION_SCOPE] if expr is DEFAULT_ACTION_SCOPE
            expr = (
                orch[name]
                if orch[name] != DEFAULT_ACTION_SCOPE
                else orch[DEFAULT_ACTION_SCOPE]
            )
            return (
                Functions.from_expr(
                    expr,
                ),
                (stop, function_response),
            )
        else:
            unavailable_function_msg = f"{name} is not a valid function name, use one of the following: {', '.join([func['name'] for func in functions.functions])}"
            messages += [
                {
                    "role": "user",
                    "content": unavailable_function_msg,
                }
            ]
            return functions, (False, None)

    def create(
        self,
        orch=None,
        actions: List[Action] = [],
        *args,
        **kwargs,
    ):
        if "model" not in kwargs:
            kwargs["model"] = self.model

        return ChatCompletion.wrap_chat_completion_create(
            self.client.chat.completions.create
        )(
            actions=actions,
            orch=orch,
            logger=self.logger,
            token_usage_tracker=self.token_usage_tracker,
            *args,
            **kwargs,
        )

    @staticmethod
    def validate_orch(orch):
        if orch is not None:
            for key in orch.keys():
                if not isinstance(key, str):
                    raise ChatCompletionException(
                        f"Orch keys must be action name (str), found {type(key)}"
                    )

    @staticmethod
    def wrap_chat_completion_create(original_create_method):
        def wrapper_for_logging(
            *args,
            logger: Optional[logging.Logger] = None,
            logging_name: Optional[str] = None,
            logging_metadata: Optional[dict] = None,
            logging_level=logging.INFO,
            **kwargs,
        ):
            DEFAULT_LOGGING_NAME = "actionweaver_initial_chat_completion"

            def new_create(
                actions: List[Action] = [],
                orch=None,
                token_usage_tracker=None,
                *args,
                **kwargs,
            ):
                ChatCompletion.validate_orch(orch)

                chat_completion_create_method = original_create_method
                if logger:
                    chat_completion_create_method = traceable(
                        name=(logging_name or DEFAULT_LOGGING_NAME)
                        + ".chat.completions.create",
                        logger=logger,
                        metadata=logging_metadata,
                        level=logging_level,
                    )(original_create_method)

                if token_usage_tracker is None:
                    token_usage_tracker = TokenUsageTracker()

                messages = kwargs.get("messages")
                if messages is None:
                    raise ChatCompletionException(
                        "messages keyword argument is required for chat completion"
                    )
                model = kwargs.get("model")
                if model is None:
                    raise ChatCompletionException(
                        "model keyword argument is required for chat completion"
                    )

                action_handler, orch = ChatCompletion.build_orch(actions, orch)

                response = None
                functions = Functions.from_expr(orch[DEFAULT_ACTION_SCOPE])

                while True:
                    function_argument = functions.to_arguments()
                    if function_argument["functions"]:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                            **function_argument,
                        )
                    else:
                        api_response = chat_completion_create_method(
                            *args,
                            **kwargs,
                        )

                    # logic to handle streaming API response
                    processed_stream_response = None
                    if isinstance(api_response, Stream):
                        processed_stream_response = (
                            ChatCompletion._handle_stream_response(api_response, logger)
                        )

                        if type(processed_stream_response) == itertools.chain:
                            # if it's a tee object, return the message right away
                            return processed_stream_response
                    else:
                        token_usage_tracker.track_usage(api_response.usage)

                    if processed_stream_response is not None:
                        functions, (
                            stop,
                            resp,
                        ) = ChatCompletion._invoke_function(
                            messages,
                            model,
                            processed_stream_response,
                            functions,
                            orch,
                            action_handler,
                        )
                        if stop:
                            return resp
                    else:
                        choice = api_response.choices[0]
                        message = choice.message

                        if message.function_call:
                            functions, (
                                stop,
                                resp,
                            ) = ChatCompletion._invoke_function(
                                messages,
                                model,
                                message.function_call,
                                functions,
                                orch,
                                action_handler,
                            )
                            if stop:
                                return resp
                        elif message.content is not None:
                            response = api_response

                            # ignore last message in the function loop
                            # messages += [{"role": "assistant", "content": message["content"]}]
                            if choice.finish_reason == "stop":
                                """
                                Stop Reasons:

                                - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                                See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                                """
                                break
                        else:
                            raise ChatCompletionException(
                                f"Unsupported response from OpenAI api: {api_response}"
                            )
                return response

            if logger:
                return traceable(
                    name=logging_name or DEFAULT_LOGGING_NAME,
                    logger=logger,
                    metadata=logging_metadata,
                    level=logging_level,
                )(new_create)(*args, **kwargs)
            else:
                return new_create(*args, **kwargs)

        return wrapper_for_logging

    @staticmethod
    def patch(client: Union[AzureOpenAI, AsyncAzureOpenAI]):
        if isinstance(client, AsyncAzureOpenAI):
            raise NotImplementedError(
                "AsyncAzureOpenAI client is not supported for patching yet."
            )

        client.chat.completions.create = ChatCompletion.wrap_chat_completion_create(
            client.chat.completions.create
        )
        return client



from typing import Union

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from actionweaver.llms.azure.chat import ChatCompletion
from actionweaver.llms.openai.tools.chat import OpenAIChatCompletion


def patch(client: Union[OpenAI, AsyncOpenAI, AsyncAzureOpenAI, AzureOpenAI]):
    if type(client) in (AsyncAzureOpenAI, AsyncOpenAI):
        raise NotImplementedError(
            "AsyncOpenAI client is not supported for patching yet."
        )
    elif type(client) == OpenAI:
        return OpenAIChatCompletion.patch(client)
    elif type(client) == AzureOpenAI:
        return ChatCompletion.patch(client)
    else:
        raise TypeError(
            f"Client type {type(client)} is not supported for patching yet."
        )



from typing import List

from pydantic import BaseModel

from actionweaver.actions.factories.function import action


class Place(BaseModel):
    lat: float
    lng: float
    description: str


class Folium:
    def verify_lib_installed(self):
        try:
            import folium
        except ImportError:
            raise ImportError(
                "`folium` package not found, please run `pip install folium`"
            )

    @action(name="ShowMap")
    def show_map(self, places: List[Place]) -> str:
        """
        Display a map with the provided latitude and longitude coordinates.

        This action requires proper integration with map services and may require API keys or authentication.
        Ensure that relevant documentation and dependencies are properly set up.

        This method requires `folium` installed
        """
        self.verify_lib_installed()

        import folium

        # Calculate the center of all places
        avg_lat = sum(place["lat"] for place in places) / len(places)
        avg_lng = sum(place["lng"] for place in places) / len(places)

        # Create a folium Map centered at the average latitude and longitude
        m = folium.Map(location=[avg_lat, avg_lng])

        # Add markers for each place
        for place in places:
            folium.Marker(
                [place["lat"], place["lng"]], tooltip=place["description"]
            ).add_to(m)

        # Display the map
        display(m)
        return "Map has been displayed"



from typing import Any, Callable, Dict, Optional, Union

ExtractorType = Callable[[str], Dict[str, Any]]


class ActionProcessor:
    def __init__(
        self,
        tools: Optional[Dict[str, Any]] = None,
        custom_extractor: Optional[ExtractorType] = None,
    ) -> None:
        self.tools = tools or {}
        self.dict = {tool.name: tool for tool in tools}
        self.custom_extractor = custom_extractor

    def extract_function(self, text: str) -> Union[Dict[str, Any], None]:
        if self.custom_extractor:
            extracted = self.custom_extractor(text)
            if (
                not isinstance(extracted, dict)
                or "name" not in extracted
                or "parameters" not in extracted
            ):
                raise ValueError(
                    "Custom extractor must return a dictionary with 'name' and 'parameters' keys."
                )
            return extracted
        else:
            import json

            j = json.loads(text)
            return {"name": j["function"], "parameters": j["parameters"]}

    def respond(self, text: str):
        function = None
        try:
            function = self.extract_function(text)
        except Exception as e:
            exception_type = type(e).__name__
            exception_message = str(e)
            full_exception_string = f"{exception_type}: {exception_message}"
            return (
                None,
                False,
                f"Unable to extract a valid function from the input. Error encountered in extractor: {full_exception_string}",
            )

        if function["name"] not in self.dict:
            return None, False, "Function or tool not found"

        response = ""
        try:
            response = self.dict[function["name"]](**function["parameters"])
        except Exception as e:
            exception_type = type(e).__name__
            exception_message = str(e)
            full_exception_string = f"{exception_type}: {exception_message}"
            return (
                None,
                False,
                f"Unable to invoke valid function {function['name']}, parameters: {function['parameters']}. Error encountered: {full_exception_string}",
            )

        return response, True, None



import inspect
import typing
from inspect import getfullargspec
from typing import Any, Callable, Dict, List, Type

from pydantic import BaseModel, create_model
from pydantic.config import ConfigDict


def create_pydantic_model_from_func_v0(
    func: Callable,
    model_name: str,
    base_model: Type[BaseModel] = BaseModel,
    config: ConfigDict = None,
    validators: dict[str, classmethod] = None,
    nested_models=None,  # models: Optional pydantic models needed for the pydantic model from function signature
    override_params=None,  # override_params: Optional dictionary of parameters to override kwarg and non-kwarg.
    ignored_params=None,  # ignored_params: Optional list of parameters to ignore.
):
    """
    This function is inspired by https://github.com/pydantic/pydantic/issues/1391


    Documentation:
    -------------
    1. https://docs.pydantic.dev/latest/concepts/models/#dynamic-model-creation

    """
    if nested_models is None:
        nested_models = []

    """
    Retrieve function signature details
    
    - args is a list of the positional parameter names.
    - varargs is the name of the * parameter or None if arbitrary positional arguments are not accepted.
    - varkw is the name of the ** parameter or None if arbitrary keyword arguments are not accepted.
    - defaults is an n-tuple of default argument values corresponding to the last n positional parameters, or None if there are no such defaults defined.
    - kwonlyargs is a list of keyword-only parameter names in declaration order.
    - kwonlydefaults is a dictionary mapping parameter names from kwonlyargs to the default values used if no argument is supplied.
    - annotations is a dictionary mapping parameter names to annotations.
    """
    (
        args,
        varargs,
        varkw,
        defaults,
        kwonlyargs,
        kwonlydefaults,
        annotations,
    ) = getfullargspec(func)

    if len(args) > 0 and args[0] == "self":
        args = args[1:]  # skip self

    defaults = defaults or []
    args = args or []

    # Calculate the number of arguments that don't have default values
    non_default_args = len(args) - len(defaults)

    # Add placeholders for non-default arguments
    defaults = (...,) * non_default_args + tuple(defaults)

    # Create a dictionary for keyword-only arguments with their defaults
    keyword_only_params = {
        param: kwonlydefaults.get(param, Any) for param in kwonlyargs
    }

    """ Create a dictionary for all arguments with their annotations and defaults
    e.g. 
    {
        'name': ('str', 'Default Name'),
        'age': ('int', 0),
        'email': ('str', None),
    }
    """

    params = {}

    if override_params is None:
        models_dict = {model.__name__: model for model in nested_models}
        for param, default in zip(args, defaults):
            annotation = annotations.get(param, Any)
            annotation = models_dict.get(annotation, annotation)
            params[param] = (annotation, default)
    else:
        # use override_params instead
        params = override_params
        keyword_only_params = {}

    ## Remove ignored params
    if ignored_params:
        params = {
            key: value for key, value in params.items() if key not in ignored_params
        }
        keyword_only_params = {
            key: value
            for key, value in keyword_only_params.items()
            if key not in ignored_params
        }

    # # Configure the class to allow extra parameters if **kwargs is in the function signature
    # class Config:
    #     extra = "allow"

    if config:
        return create_model(
            model_name,
            **params,
            **keyword_only_params,
            __config__=config,
            __validators__=validators,
        )
    else:
        return create_model(
            model_name,
            **params,
            **keyword_only_params,
            __base__=base_model,
            __validators__=validators,
        )


def convert_default(val):
    default_mapping = {inspect._empty: Ellipsis}
    return default_mapping.get(val, val)


def convert_annotation(val):
    annotation_mapping = {inspect._empty: Any}
    return annotation_mapping.get(val, val)


def create_pydantic_model_from_func(
    model_name: str,
    func: Callable,
    base_model: Type[BaseModel] = BaseModel,
    config: ConfigDict = None,
    validators: Dict[str, classmethod] = None,
    override_params: Dict[str, Any] = None,
    ignored_params: List[str] = None,
):
    """
    Create a Pydantic model from a function signature.

    If inspect signature return string for imported methods, consider removing `from __future__ import annotations`
    Documentation:
    -------------
    1. https://docs.pydantic.dev/latest/concepts/models/#dynamic-model-creation
    2. https://github.com/pydantic/pydantic/issues/1391
    """

    # Retrieve function signature details using inspect.signature
    signature = inspect.signature(func)
    parameters = list(signature.parameters.values())

    # Filter out 'self' if present
    params = {param.name: param for param in parameters if param.name != "self"}

    # Convert to a dictionary of parameter names and their annotations and defaults
    params = {
        name: (convert_annotation(param.annotation), convert_default(param.default))
        for name, param in params.items()
    }

    # Use override_params instead
    if override_params:
        params = override_params

    # Remove ignored params
    if ignored_params:
        params = {
            key: value for key, value in params.items() if key not in ignored_params
        }

    # (Deprecated) Convert annotations to pydantic models if needed
    # if supporting_annotation_models:
    #     models_dict = {model.__name__: model for model in supporting_annotation_models}
    #     for name, (annotation, default) in params.items():
    #         annotation = models_dict.get(annotation, annotation)
    #         params[name] = (annotation, default)

    if config:
        return create_model(
            model_name,
            **params,
            __config__=config,
            __validators__=validators,
        )
    else:
        return create_model(
            model_name,
            **params,
            __base__=base_model,
            __validators__=validators,
        )



from typing import Union

from openai import AzureOpenAI, OpenAI

from actionweaver.llms.azure.chat_loop import create_chat_loop as create_chat_loop_azure
from actionweaver.llms.openai.tools.chat_loop import create_chat_loop


class ActionWeaverLLMClientWrapper:
    def __init__(self, client: Union[OpenAI, AzureOpenAI]):

        self.client = client
        if type(client) == OpenAI:
            self.chat_loop = create_chat_loop(client.chat.completions.create)
        elif type(client) == AzureOpenAI:
            self.chat_loop = create_chat_loop_azure(client.chat.completions.create)
        else:
            raise NotImplementedError(f"Client type {type(client)} is not supported.")

    def create(self, *args, **kwargs):
        return self.chat_loop(*args, **kwargs)


def wrap(client: Union[OpenAI, AzureOpenAI]):
    return ActionWeaverLLMClientWrapper(client)



# from openai import OpenAI

# client = OpenAI()

# from actionweaver import action


# class OpenAIAPI:
#     @action(name="GenerateImage")
#     def generate_image(self, prompt: str) -> str:
#         """
#         Generates an image based on a given prompt.

#         Args:
#             prompt (str): A descriptive text input that guides the content of the generated image (e.g., "lion running on Mars").

#         Returns:
#             str: A message indicating that the image has been created, along with the URL of the generated image.
#         """
#         from IPython.display import Image, display

#         response = client.images.generate(prompt=prompt, n=1, size="1024x1024")

#         url = response.data[0].url
#         display(Image(url=url))
#         return "The image has been created"
